#!/usr/bin/env python3
"""
Virus Data Retrieval Module for Dane County School Dashboard

This module retrieves virus surveillance data from LabKey using the results_summary_qualitative
table with academic year filtering and assay system precedence. It supports filtering by
academic year periods and applies assay system precedence rules.

Usage as a module:
    from retrieve_virus_data import retrieve_school_virus_data, filter_by_academic_year
    
    # Get school virus data for current academic year
    df = retrieve_school_virus_data()
    
    # Get data for specific academic year
    df_2024 = retrieve_school_virus_data(academic_year="2024-2025")
"""

import sys
import os
from pathlib import Path
from typing import Dict, List, Any, Optional, Union
import pandas as pd
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
import pickle
import gzip
from datetime import datetime

# Load environment variables from .env file if running locally
try:
    from dotenv import load_dotenv
    # Look for .env file in the project root
    env_path = Path(__file__).parent.parent.parent / '.env'
    if env_path.exists():
        load_dotenv(env_path)
except ImportError:
    # dotenv not available, rely on environment variables being set externally
    pass

# Add the labkey_api module to the path
sys.path.insert(0, str(Path(__file__).parent.parent.parent / 'labkey_api'))

from utils.labkey_client import LabKeyClient
from academic_year_utils import get_current_academic_year, get_academic_year_date_range, filter_by_academic_year
from assay_precedence import AssayPrecedenceManager


def get_essential_fields_qualitative() -> List[str]:
    """
    Get the essential fields needed from results_summary_qualitative table.
    
    These are the minimum fields required for school virus processing pipeline.
    Based on actual table schema discovered via debugging.
    
    Returns:
        List of field names to retrieve from LabKey
    """
    return [
        'facility_name',      # School facility name
        'start_time',         # Sample start time (used as sample_date)
        'end_time',           # Sample end time
        'assay_system',       # Assay system used
        'assay_target',       # Virus/pathogen target
        'qualitative_result', # Qualitative result (positive/negative/etc)
        'average_result',     # Quantitative result (average)
        'copies_per_reaction', # Copies per reaction if available
        'result_reported',    # Date when result was reported
        'program_id',         # Program identifier
        'cartridge_id',       # Cartridge identifier
        'detection_assay',    # Detection assay details
        'sampler_name'        # Sampler name
    ]


def create_school_virus_query(facilities: List[str] = None,
                             assay_systems: List[str] = None,
                             academic_year: str = None,
                             max_rows: Optional[int] = None,
                             test_filter: bool = False,
                             oregon_test: bool = False) -> str:
    """
    Create an optimized SQL query to retrieve school virus data.
    
    Args:
        facilities: List of facility names to include (None for all schools)
        assay_systems: List of assay systems to include (None for all enabled)
        academic_year: Academic year to filter by (e.g., "2024-2025")
        max_rows: Maximum number of rows to return (None for unlimited)
        test_filter: If True, add WHERE clause for limited test data
        oregon_test: If True, filter to Oregon High School data from Aug 2024 - Jun 2025
        
    Returns:
        SQL query string for results_summary_qualitative table
    """
    # Use essential fields for efficiency
    essential_fields = get_essential_fields_qualitative()
    fields_str = ', '.join(essential_fields)
    
    # Base table for school surveillance
    base_table = 'results_summary_qualitative'
    
    # Build WHERE conditions
    conditions = []
    
    # Filter for school facilities (contain "School" in name)
    conditions.append("LOWER(facility_name) LIKE '%school%'")
    
    # Filter by specific facilities if provided
    if facilities:
        facility_conditions = [f"facility_name = '{facility}'" for facility in facilities]
        conditions.append(f"({' OR '.join(facility_conditions)})")
    
    # Filter by assay systems if provided, otherwise default to BioMarkX only
    if assay_systems:
        assay_conditions = [f"assay_system = '{assay}'" for assay in assay_systems]
        conditions.append(f"({' OR '.join(assay_conditions)})")
    else:
        # Default to BioMarkX Respiratory Pathogen Panel 001 only
        conditions.append("assay_system = 'BioMarkX Respiratory Pathogen Panel 001'")
    
    # Add basic data quality filters
    conditions.extend([
        "facility_name IS NOT NULL",
        "facility_name != ''",
        "start_time IS NOT NULL",
        "assay_target IS NOT NULL",
        "assay_target != ''"
    ])
    
    # Apply date filters based on test type
    if oregon_test:
        # Filter to Oregon High School data from Aug 2024 - Jun 2025 (2024-25 academic year)
        conditions.append("facility_name = 'Oregon High School'")
        conditions.append("start_time >= '2024-08-15'")  # Aug 15, 2024
        conditions.append("start_time <= '2025-06-01'")  # Jun 1, 2025
    elif test_filter:
        # Filter to November 1 - February 1 date range from two schools with consistent data
        # Using November 2023 - February 2024 period to ensure we have actual historical data
        test_conditions = [
            "start_time >= '2023-11-01'",  # November 1, 2023
            "start_time <= '2024-02-29'"   # February 29, 2024 (leap year)
        ]
        conditions.extend(test_conditions)
    
    # Combine conditions
    where_clause = ' AND '.join(conditions)
    
    # Build final query
    sql = f"SELECT {fields_str} FROM {base_table} WHERE {where_clause}"
    
    # Add ordering for consistency
    sql += " ORDER BY facility_name, start_time DESC, assay_target"
    
    # Add limit if specified
    if max_rows:
        sql += f" LIMIT {max_rows}"
    elif test_filter:
        sql += " LIMIT 1000"
    elif oregon_test:
        sql += " LIMIT 5000"  # Allow more data for full academic year
    
    return sql.strip()


def retrieve_school_virus_data(facilities: List[str] = None,
                              assay_systems: List[str] = None,
                              academic_year: str = None,
                              max_rows: Optional[int] = None,
                              apply_precedence: bool = True,
                              show_progress: bool = True,
                              test_filter: bool = False,
                              oregon_test: bool = False,
                              cache_file: Optional[Path] = None,
                              use_cache: bool = False) -> pd.DataFrame:
    """
    Retrieve school virus surveillance data from results_summary_qualitative table.
    
    Args:
        facilities: List of facility names to include (None for all schools)
        assay_systems: List of assay systems to include (None for all enabled)
        academic_year: Academic year to filter by (None for current academic year)
        max_rows: Maximum number of rows to return (None for unlimited)
        apply_precedence: Whether to apply assay precedence rules
        show_progress: Whether to show progress indicators
        test_filter: If True, filter to recent data for fast testing
        oregon_test: If True, filter to Oregon High School Aug 2024 - Jun 2025
        cache_file: Path to cache file for saving/loading data
        use_cache: If True and cache exists, load from cache instead of querying
        
    Returns:
        pandas DataFrame with school virus surveillance data
    """
    console = Console()
    
    # Use current academic year if none specified
    if academic_year is None:
        academic_year = get_current_academic_year()
    
    # Try to load from cache if requested
    if use_cache and cache_file and cache_file.exists():
        if show_progress:
            console.print(f"[cyan]Loading school virus data from cache: {cache_file}[/cyan]")
        
        try:
            df, cache_info = load_from_cache(cache_file)
            age_days = (datetime.now() - cache_info['timestamp']).days
            
            if show_progress:
                console.print(f"[green]✓ Loaded {len(df)} records from cache[/green]")
                console.print(f"  Cache age: {age_days} days old")
                console.print(f"  Academic year: {cache_info.get('academic_year', 'Unknown')}")
                
                if age_days > 7:
                    console.print(f"[yellow]⚠ Cache is {age_days} days old. Consider refreshing.[/yellow]")
            
            return df
            
        except Exception as e:
            console.print(f"[yellow]Failed to load cache: {e}[/yellow]")
            console.print("[cyan]Falling back to fresh data retrieval...[/cyan]")
    
    try:
        # Initialize LabKey client
        base_path = Path(__file__).parent
        labkey_client = LabKeyClient(base_path)
        
        if show_progress:
            filter_msg = f" for {academic_year}" if academic_year else ""
            test_msg = " (test mode)" if test_filter else ""
            oregon_msg = " (Oregon High School 2024-25)" if oregon_test else ""
            console.print(f"[blue]Retrieving school virus data from LabKey{filter_msg}{test_msg}{oregon_msg}...[/blue]")
        
        # Create SQL query
        sql = create_school_virus_query(
            facilities=facilities,
            assay_systems=assay_systems,
            academic_year=academic_year,
            max_rows=max_rows,
            test_filter=test_filter,
            oregon_test=oregon_test
        )
        
        # Debug: print SQL query
        if show_progress:
            console.print(f"[dim]Generated SQL: {sql}[/dim]")
        
        # Execute query with progress tracking
        if show_progress:
            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                TimeElapsedColumn(),
                console=console,
            ) as progress:
                task = progress.add_task("Executing query...", total=None)
                
                try:
                    result = labkey_client.execute_sql(schema_name="lists", sql=sql)
                    df = pd.DataFrame(result.get('rows', []))
                    progress.update(task, description=f"Retrieved {len(df)} records", completed=100)
                except Exception as e:
                    progress.update(task, description=f"Query failed", completed=100)
                    raise e
        else:
            # Silent retrieval
            result = labkey_client.execute_sql(schema_name="lists", sql=sql)
            df = pd.DataFrame(result.get('rows', []))
        
        # Basic data cleaning
        if not df.empty:
            # Convert date columns to datetime
            date_columns = ['start_time', 'end_time', 'result_reported']
            for col in date_columns:
                if col in df.columns:
                    df[col] = pd.to_datetime(df[col], errors='coerce')
            
            # Convert numeric columns
            numeric_columns = ['average_result', 'copies_per_reaction']
            for col in numeric_columns:
                if col in df.columns:
                    df[col] = pd.to_numeric(df[col], errors='coerce')
            
            # Add a unified sample_date column using start_time for compatibility
            if 'start_time' in df.columns:
                df['sample_date'] = df['start_time']
            
            # Add a unified target_name column using assay_target for compatibility
            if 'assay_target' in df.columns:
                df['target_name'] = df['assay_target']
            
            # Remove any completely null rows
            df = df.dropna(how='all')
            
            # Apply assay precedence rules if requested
            if apply_precedence and 'assay_system' in df.columns:
                try:
                    assay_manager = AssayPrecedenceManager()
                    df = assay_manager.apply_precedence_rules(df)
                    if show_progress:
                        console.print("[cyan]✓ Applied assay precedence rules[/cyan]")
                except Exception as e:
                    if show_progress:
                        console.print(f"[yellow]Warning: Could not apply assay precedence: {e}[/yellow]")
        
        if show_progress:
            console.print(f"[green]✓ Successfully retrieved {len(df)} school virus records[/green]")
            if academic_year:
                console.print(f"  Academic year: {academic_year}")
            if 'facility_name' in df.columns:
                unique_schools = df['facility_name'].nunique()
                console.print(f"  Schools: {unique_schools}")
        
        # Save to cache if cache_file is specified
        if cache_file and not df.empty:
            if show_progress:
                console.print(f"[cyan]Saving data to cache: {cache_file}[/cyan]")
            
            # Create enhanced metadata
            metadata = {
                'academic_year': academic_year,
                'facilities': facilities,
                'assay_systems': assay_systems,
                'max_rows': max_rows,
                'test_filter': test_filter,
                'apply_precedence': apply_precedence,
                'row_count': len(df),
                'columns': list(df.columns),
                'unique_schools': df['facility_name'].nunique() if 'facility_name' in df.columns else 0
            }
            
            # Add date range information
            if not df.empty and 'start_time' in df.columns:
                metadata['latest_sample_date'] = df['start_time'].max()
                metadata['earliest_sample_date'] = df['start_time'].min()
            
            try:
                save_to_cache(df, cache_file, metadata)
                if show_progress:
                    console.print(f"[green]✓ Saved {len(df)} records to cache[/green]")
            except Exception as e:
                console.print(f"[yellow]Failed to save cache: {e}[/yellow]")
        
        return df
        
    except Exception as e:
        error_msg = f"Failed to retrieve school virus data: {e}"
        if show_progress:
            console.print(f"[red]✗ {error_msg}[/red]")
        raise Exception(error_msg)


def save_to_cache(df: pd.DataFrame, cache_file: Path, metadata: Dict[str, Any]) -> None:
    """Save DataFrame to compressed pickle with metadata."""
    cache_data = {
        'data': df,
        'metadata': metadata,
        'timestamp': datetime.now(),
        'version': '1.0'
    }
    
    cache_file.parent.mkdir(parents=True, exist_ok=True)
    
    with gzip.open(cache_file, 'wb') as f:
        pickle.dump(cache_data, f, protocol=pickle.HIGHEST_PROTOCOL)


def load_from_cache(cache_file: Path) -> tuple[pd.DataFrame, Dict[str, Any]]:
    """Load DataFrame from compressed pickle with metadata."""
    with gzip.open(cache_file, 'rb') as f:
        cache_data = pickle.load(f)
    
    return cache_data['data'], cache_data


def get_available_schools(df: pd.DataFrame = None, min_samples: int = 10) -> List[str]:
    """
    Get list of schools with sufficient data.
    
    Args:
        df: DataFrame to analyze (if None, queries LabKey directly)
        min_samples: Minimum number of samples required
        
    Returns:
        List of school facility names with sufficient data
    """
    if df is None:
        # Query LabKey for school summary
        sql = """
        SELECT 
            facility_name,
            COUNT(*) as sample_count
        FROM results_summary_qualitative 
        WHERE LOWER(facility_name) LIKE '%school%'
        AND facility_name IS NOT NULL 
        AND facility_name != ''
        GROUP BY facility_name
        HAVING COUNT(*) >= {}
        ORDER BY sample_count DESC
        """.format(min_samples)
        
        try:
            base_path = Path(__file__).parent
            labkey_client = LabKeyClient(base_path)
            result = labkey_client.execute_sql(schema_name="lists", sql=sql)
            df_summary = pd.DataFrame(result.get('rows', []))
            
            if not df_summary.empty:
                return df_summary['facility_name'].tolist()
            else:
                return []
                
        except Exception as e:
            print(f"Error querying available schools: {e}")
            return []
    
    else:
        # Analyze provided DataFrame
        if 'facility_name' not in df.columns:
            return []
        
        school_counts = df['facility_name'].value_counts()
        sufficient_schools = school_counts[school_counts >= min_samples]
        return sufficient_schools.index.tolist()


def get_data_summary(df: pd.DataFrame, academic_year: str = None) -> Dict[str, Any]:
    """
    Get summary statistics for the retrieved school virus data.
    
    Args:
        df: pandas DataFrame with school virus data
        academic_year: Academic year for context
        
    Returns:
        Dictionary with summary statistics
    """
    if df.empty:
        return {"total_records": 0, "academic_year": academic_year, "message": "No data available"}
    
    summary = {
        "total_records": len(df),
        "academic_year": academic_year,
        "date_range": {
            "earliest": df['sample_date'].min() if 'sample_date' in df.columns else None,
            "latest": df['sample_date'].max() if 'sample_date' in df.columns else None
        },
        "unique_schools": df['facility_name'].nunique() if 'facility_name' in df.columns else 0,
        "unique_targets": df['assay_target'].nunique() if 'assay_target' in df.columns else 0,
        "unique_assays": df['assay_system'].nunique() if 'assay_system' in df.columns else 0,
        "columns": list(df.columns)
    }
    
    # Add school-specific statistics
    if 'facility_name' in df.columns:
        school_counts = df['facility_name'].value_counts()
        summary["schools"] = {
            "total": len(school_counts),
            "top_5": school_counts.head().to_dict(),
            "samples_per_school_avg": school_counts.mean(),
            "samples_per_school_median": school_counts.median()
        }
    
    # Add assay system statistics
    if 'assay_system' in df.columns:
        assay_counts = df['assay_system'].value_counts()
        summary["assay_systems"] = {
            "total": len(assay_counts),
            "distribution": assay_counts.to_dict(),
            "most_common": assay_counts.index[0] if len(assay_counts) > 0 else None
        }
    
    # Add target statistics
    if 'assay_target' in df.columns:
        target_counts = df['assay_target'].value_counts()
        summary["targets"] = {
            "total": len(target_counts),
            "top_10": target_counts.head(10).to_dict()
        }
    
    return summary


def main():
    """Demonstration of school virus data retrieval functionality."""
    console = Console()
    
    console.print("[bold green]🏫 School Virus Data Retrieval Module[/bold green]\n")
    
    # Check if API key is available
    if not os.environ.get('LABKEY_API'):
        console.print("[yellow]⚠ LABKEY_API environment variable not set[/yellow]")
        console.print("Set your LabKey API key to test data retrieval:")
        console.print("export LABKEY_API='your_api_key_here'")
        return 1
    
    try:
        # Test 1: Get available schools
        console.print("[blue]Test 1: Getting available schools...[/blue]")
        schools = get_available_schools(min_samples=5)
        console.print(f"✓ Found {len(schools)} schools with sufficient data")
        if schools:
            console.print(f"Sample schools: {schools[:3]}")
        
        # Test 2: Retrieve current academic year data (sample)
        console.print("\n[blue]Test 2: Retrieving current academic year data (10 records sample)...[/blue]")
        current_year = get_current_academic_year()
        sample_data = retrieve_school_virus_data(
            academic_year=current_year,
            max_rows=10,
            test_filter=True
        )
        
        console.print(f"✓ Retrieved {len(sample_data)} records for {current_year}")
        if not sample_data.empty:
            console.print(f"Columns: {list(sample_data.columns)}")
            if 'facility_name' in sample_data.columns:
                console.print(f"Sample schools: {sample_data['facility_name'].unique()[:3].tolist()}")
        
        # Test 3: Get data summary
        if not sample_data.empty:
            console.print(f"\n[blue]Test 3: Data summary for {current_year}...[/blue]")
            summary = get_data_summary(sample_data, current_year)
            
            for key, value in summary.items():
                if key not in ["columns", "schools", "assay_systems", "targets"]:
                    console.print(f"  {key}: {value}")
        
        console.print("\n[green]✅ All tests completed successfully![/green]")
        console.print("\n[cyan]Primary usage as a module:[/cyan]")
        console.print("from retrieve_virus_data import retrieve_school_virus_data")
        console.print("df = retrieve_school_virus_data()  # Gets current academic year data")
        console.print("df_2024 = retrieve_school_virus_data(academic_year='2024-2025')")
        
        return 0
        
    except Exception as e:
        console.print(f"[red]❌ Error: {e}[/red]")
        import traceback
        traceback.print_exc()
        return 1


if __name__ == "__main__":
    exit_code = main()
    sys.exit(exit_code)